#!/usr/bin/perl -w

#===============================================================================
# Single-end (SE) read sequence preprocessing, specific for "4N adapters" on 3'-end only
# 1) use Cutadapt for 3' adapter trmming and 5' adatper trmming
# 2) use Cutadapt for empty read (only adapter but no insert sequence) discarding
# 3) use Prinseq for low-quality read filtering
# 4) use FASTX_collapser for identical read collapsing
# 5) use Cutadapt for stop oligo filtering (optional)
# mainly for Illumina TruSeq Small RNA Sample Prep Kit plus 4 random nucleotides (4N) on 3'-end only: 
# 	Illumina TruSeq Small RNA Sample Prep Kit
# 		3’ Adapter: TGGAATTCTCGGGTGCCAAG
#		e.g., TGGAATTCTCGGGTGCCAAG GAACTCCAGTCAC XXXXXX ATCTCGTATGCC GTCTTCTGCTTG
#									   barcode - XXXXXX
# 		5’ Adapter:	GTTCAGAGTTCTACAGTCCGACGATC
#		Stop Oligo:	GAATTCCACCA CGTTCCCG TGG
#					        CCATCGTTCCCGATGG
#            		             GTTCCCGATGG
#===============================================================================

use strict;
use warnings;

my $option_01 = $ARGV[0];
if ( !$ARGV[0] ) {
	$option_01 = "-z";
}

my $option_02 = $ARGV[1];
if ( !$ARGV[1] ) {
	$option_02 = "--stp";
}

my $dataDIR = "./";

opendir my $dir, $dataDIR or die "Cannot open directory: $dataDIR";
my @fileList = readdir $dir;
closedir $dir;

my $suffix = "\.fastq";
if ( $option_01 eq "-z" ) {
	$suffix = "\.fastq.gz";
}
my @readFileList = grep(/$suffix$/, @fileList);
@readFileList = sort @readFileList;

my $num = scalar(@readFileList);
if ( $num == 0 ) {
	print "\nThere is NO FASTQ file ($suffix) in this folder\n";
	exit;
} elsif ( $num == 1 ) {
	print "\nThere is ".$num." FASTQ file ($suffix) in this folder\n";
} else {
	print "\nThere are ".$num." FASTQ file ($suffix) in this folder\n";
}

# parameters for cutadapt
my $adapter_3 = "NNNNTGGAATTCTCGGGTGCCAAGGAACTCCAGTCAC";
my $adapter_5 = "GTTCAGAGTTCTACAGTCCGACGATC";
my $emptyIn_5 = "AGAGTTCTACAGTCCGAC";
my $stop_oligo_A = "CGTTCCCGATGG\$";
my $stop_oligo_B = "CCATCGTTCCCG";
my $err_3 = 0.2;
my $err_5 = 0.125;
my $err_BC = 0.0625;
my $overlap_3 = 5;
my $overlap_5 = 8;
my $overlap_BC = 12;
my $overlap_STP = 11;
my $min_length = 1;

# parameters for prinseq
my $lc_method = "entropy";
my $lc_threshold = 0;
my $min_len = 1;
my $trim_tail_right = 1;
my $line_width = 86;

# read sample barcode file
my %sampleBarcode = ();
open(In, "<", "sampleBarcode") or die "\n   can't find sampleBarcode file !!!\n";
while (<In>) {
	chomp;
	my @array = split(/\t/, $_);
	if ( scalar(@array) > 1 ) {
		$sampleBarcode{$array[0]} = $array[1];
	}
}
close(In);

my $i = 0;
foreach my $readFile (@readFileList) {
	$i++;
	my $prefix = $readFile;
	$prefix =~ s/$suffix$//;
	
	my $barcode;
	if ( exists $sampleBarcode{$prefix} ) {
		$barcode = $sampleBarcode{$prefix};
	} else {
		print "\n   there's no barcode for sample: $prefix!!!\n";
		next;
	}
	
	print "\n====== Preprocess #".$i." read file: $prefix ======\n";
	preprocess($prefix, $suffix, $barcode);
}

#================================ sub functions ================================
# preprocess
sub preprocess {
	my ($prefix, $suffix, $barcode) = @_;
	
	my $readFile = $prefix.$suffix;
	
	my $barcodeArea_1 = $barcode."ATCTCGTATGCCG";
	my $barcodeArea_2 = "CAGTCAC".$barcode."ATCTCGT";
	my $barcodeArea_3 = "GAACTCCAGTCAC".$barcode;

	
	my $trimedReadFile = $prefix."\_Cutadapt.fastq";
	my $trimReport = $prefix."\_Cutadapt.report";
	
	my $filteredReadFile = $prefix."\_Prinseq";
	my $filterReport = $prefix."\_Cutadapt_Prinseq.report";
	
	my $collapsedReadFile = $prefix."\_Collapsed.fa";
	my $processedReadFile = $prefix."\_Processed.fa";
	my $stpReadFile0 = $prefix."\_stp_mis_0.fa";
	my $stpReadFile1 = $prefix."\_stp_mis_1.fa";
	my $stpReadFile2 = $prefix."\_stp_mis_2.fa";
	my $stpReport = $prefix."\_stp.report"; 
	
	# use Cutadapt for 3' adapter trmming and 5' adatper trmming
	system("cutadapt -a $adapter_3 -n 1 -e $err_3 -O $overlap_3 -m $min_length --match-read-wildcards -o tmp_Cutadapt3.fastq $readFile > $trimReport");
	system("cutadapt -g $adapter_5 -n 1 -e $err_5 -O $overlap_5 -m $min_length --match-read-wildcards -o tmp_Cutadapt.fastq tmp_Cutadapt3.fastq >> $trimReport");
	
	# use Cutadapt for empty read (only adapter but no insert sequence) discarding
	system("cutadapt -b $barcodeArea_1 -b $barcodeArea_2 -b $barcodeArea_3 -g $emptyIn_5 -n 1 -e $err_BC -O $overlap_BC -m $min_length --match-read-wildcards --discard-trimmed -o $trimedReadFile tmp_Cutadapt.fastq >> $trimReport");
	
	system("rm tmp_Cutadapt3.fastq tmp_Cutadapt.fastq");
	print "   --- Adatper trmming for $prefix completed\n";
	
	# use Prinseq for low-quality read filtering
	system("prinseq-lite.pl -fastq $trimedReadFile -out_format 1 -out_good $filteredReadFile -out_bad null -lc_method $lc_method -lc_threshold $lc_threshold -min_len $min_len -line_width $line_width -trim_tail_right $trim_tail_right 2> $filterReport");

	system("rm $trimedReadFile");
	print "   --- Low-quality read filtering for $prefix completed\n";

	# use FASTX_collapser for identical read collapsing
	system("fastx_collapser -i $filteredReadFile\.fasta -o $collapsedReadFile");
	
	system("rm $filteredReadFile\.fasta");
	print "   --- Identical read collapsing for $prefix completed\n";
	
	# use Cutadapt for stop oligo filtering
	if ($option_02 eq "--stp") {
		system("cutadapt -b $stop_oligo_A -b $stop_oligo_B -n 1 -e 0.000 -O $overlap_STP --no-trim --untrimmed-output stp0_Processed.fa -o $stpReadFile0 $collapsedReadFile >> $stpReport");	# maxMismatch = 0
		system("cutadapt -b $stop_oligo_A -n 1 -e 0.091 -O $overlap_STP --no-trim --untrimmed-output stp1_Processed.fa -o $stpReadFile1 stp0_Processed.fa >> $stpReport");						# maxMismatch = 1
		system("cutadapt -b $stop_oligo_A -n 1 -e 0.182 -O $overlap_STP --no-trim --untrimmed-output $processedReadFile -o $stpReadFile2 stp1_Processed.fa >> $stpReport");						# maxMismatch = 2
		
		system("rm $collapsedReadFile stp0_Processed.fa stp1_Processed.fa");
		print "   --- Stop oligo trimming for $prefix completed\n";
	} else {
		system("mv $collapsedReadFile $processedReadFile");
	}
}